/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2009                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/vertex/component.h>

#include <vcg/complex/used_types.h>

#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component.h>
#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_ply.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/space/index/grid_static_ptr.h>
#include<vcg/complex/algorithms/closest.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/update/bounding.h>
using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::VertexRef,  face::Normal3f, face::Mark, face::FFAdj, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};

int main( int argc, char **argv )
{
  MyMesh bm, pm, om;
  GridStaticPtr<MyFace> myGrd;

  if(argc < 3 )
  {
    printf("trimeshbase basemesh.ply pointmesh.ply\n");
    return -1;
  }

  printf("Reading %s\n",argv[1]);

  int ret= tri::io::ImporterPLY<MyMesh>::Open(bm,argv[1]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[1],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  ret = tri::io::ImporterPLY<MyMesh>::Open(pm,argv[2]);
  if(ret!=0) {
    printf("Unable to open %s for '%s'\n",argv[2],tri::io::ImporterPLY<MyMesh>::ErrorMsg(ret));
    return -1;
  }
  printf("BaseMesh has %i vert and %i faces\n",bm.vn, bm.fn);
  printf("PointMesh has %i vert and %i faces\n",pm.vn, pm.fn);
  tri::UpdateNormals<MyMesh>::PerFaceNormalized(bm);
  tri::UpdateBounding<MyMesh>::Box(bm);

  // Note for the students:
  // The embarassing bug we encountered was due to the absence of stupid initialization step.
  // that involved the precomputation of the best projection direction.
  // I have patched the library so that it is no more necessary...

  myGrd.Set(bm.face.begin(),bm.face.end(),100);

  float maxdist = bm.bbox.Diag()/100.0;
  float founddist;
  Point3f cp;

  tri::FaceTmark<MyMesh> markerFunctor(&bm);
  face::PointDistanceBaseFunctor<float> PDistFunct;

  tri::Allocator<MyMesh>::AddVertices(om,bm.vn);
  for(size_t i=0;i<bm.vn;++i)
  {
    om.vert[i].P() = bm.vert[i].cP();
  }

  tri::UpdateFlags<MyMesh>::FaceClearV(bm);
  tri::UpdateFlags<MyMesh>::VertexClearV(pm);
  for(size_t i=0;i<pm.vn;++i)
  {
    Point3f qp = pm.vert[i].cP(); // search the face nearest to the the i-th vertex of the point mesh;
    MyMesh::FacePointer fp=myGrd.GetClosest(PDistFunct,markerFunctor,qp,maxdist,founddist,cp);
    if(fp)
    {
      if(!(fp->IsV()))  // we copy and split un-visited faces onto a new mesh;
      {
        fp->SetV();
        int v0 = tri::Index(bm,fp->V(0));
        int v1 = tri::Index(bm,fp->V(1));
        int v2 = tri::Index(bm,fp->V(2));

        MyFace *fp = &*tri::Allocator<MyMesh>::AddFaces(om,3);
        tri::Allocator<MyMesh>::AddVertices(om,1);

        om.vert.back().P()=qp;
        int nv = om.vn-1;

        fp[0].V(0)=&om.vert[v0];
        fp[0].V(1)=&om.vert[v1];
        fp[0].V(2)=&om.vert[nv];

        fp[1].V(0)=&om.vert[v1];
        fp[1].V(1)=&om.vert[v2];
        fp[1].V(2)=&om.vert[nv];

        fp[2].V(0)=&om.vert[v2];
        fp[2].V(1)=&om.vert[v0];
        fp[2].V(2)=&om.vert[nv];
      }
    } else {
      // we have built our point set with point onto the surface so it should always find a correct face
      assert(0);
    }
  }

  tri::io::ExporterPLY<MyMesh>::Save(om,"out.ply");
  return 0;
}
